from unittest.mock import patch

from estraces.formats.ets_writer import _get_optimal_batch_size

from .context import estraces  # noqa
from .conftest import _HEADERS
import numpy as np
import os
from estraces import ETSWriter
from estraces import read_ths_from_ets_file
import pytest
from pathlib import Path

nb_trace = 1000
nb_points = 500
filename = 'ETS-test.ets'
compressed = 'ETS.etsz'
filename_1 = 'ETS1.ets'
filename_2 = 'ETS2.ets'
filename_3 = 'ETS3.ets'

filenames = [filename, filename_1, filename_2, filename_3, compressed]


@pytest.fixture
def ets_filenames():
    for fn in filenames:
        try:
            os.remove(fn)
        except FileNotFoundError:
            pass
    yield filenames
    for fn in filenames:
        try:
            os.remove(fn)
        except FileNotFoundError:
            pass


def test_ets_writer_raises_exception_if_trying_to_replace_existing_data_wo_overwrite_mode(ets_filenames):
    out = ETSWriter(filename)
    datas = np.random.randint(100, size=(nb_trace, nb_points), dtype=np.uint8)
    out.write_samples(datas)
    out.close()
    out = ETSWriter(filename)
    with pytest.raises(estraces.ETSWriterError):
        out.write_samples(datas, index=0)
    out.close()
    out = ETSWriter(filename, overwrite=True)
    replacement = np.random.randint(100, size=(nb_trace - 5, nb_points), dtype=np.uint8)
    out.write_samples(replacement, index=0)
    ths = out.get_reader()
    assert np.array_equal(ths.samples[:nb_trace - 5], replacement)


def test_ets_writer_works_whatever_is_the_index_order(ets_filenames):
    base_trace = np.random.randint(0, 256, (nb_trace, nb_points), dtype='uint8')

    ets1 = ETSWriter(filename_1, overwrite=True)
    indexes = np.arange(nb_trace)
    for ind in indexes:
        ets1.write_samples(base_trace[ind], index=ind)
    ets1.close()

    ets2 = ETSWriter(filename_2, overwrite=True)

    for ind in reversed(indexes):
        ets2.write_samples(base_trace[ind], index=ind)
    ets2.close()

    ets3 = ETSWriter(filename_3, overwrite=True)
    indexes_3 = [nb_trace - 1, 2, 1]
    for ind in indexes_3:
        ets3.write_samples(base_trace[ind], index=ind)
    ets3.close()

    d1 = read_ths_from_ets_file(filename_1).samples[:]
    d2 = read_ths_from_ets_file(filename_2).samples[:]
    d3 = read_ths_from_ets_file(filename_3).samples[:]
    assert np.array_equal(d1[nb_trace - 1], d3[nb_trace - 1])
    assert np.array_equal(d1[2], d3[2])
    assert np.array_equal(d1[1], d3[1])

    assert d1.shape == d2.shape
    assert d1.shape == d3.shape

    s1 = os.path.getsize(filename_1)
    s2 = os.path.getsize(filename_2)

    assert s1 == s2


def test_write_ndarray(ets_filenames):
    out = ETSWriter(filename, overwrite=True)

    datas = np.random.randint(100, size=(nb_trace, nb_points), dtype=np.uint8)
    plaintext = np.random.randint(256, size=(nb_trace, 16), dtype=np.uint8)
    for index, data in enumerate(datas):
        out.write_samples(data, index=index)
        out.write_metadata('plaintext', plaintext[index], index=index)

    ths = out.get_reader()
    for i, t in enumerate(ths):
        assert np.array_equal(t.samples[:], datas[i])
        assert np.array_equal(t.plaintext, plaintext[i])


def test_write_2d_ndarray_metadata(ets_filenames):
    out = ETSWriter(filename, overwrite=True)

    datas = np.random.randint(100, size=(nb_trace, nb_points), dtype=np.uint8)
    plaintext = np.random.randint(256, size=(nb_trace, 16), dtype=np.uint8)
    for index, data in enumerate(datas):
        out.write_samples(data, index=index)
        out.write_metadata('plaintext', plaintext[index][None, :], index=index)

    ths = out.get_reader()
    for i, t in enumerate(ths):
        assert np.array_equal(t.samples[:], datas[i])
        assert np.array_equal(t.plaintext, plaintext[i])


def test_write_2d_ndarray_points(ets_filenames):
    out = ETSWriter(filename, overwrite=True)
    datas = np.random.randint(100, size=(nb_trace, nb_points), dtype=np.uint8)
    out.write_samples(datas)
    ths = out.get_reader()
    assert np.array_equal(ths.samples.array, datas)


def test_write_points_length_1(ets_filenames):
    out = ETSWriter(filename, overwrite=True)

    datas = np.random.randint(100, size=(nb_trace, nb_points), dtype=np.uint8)
    for index, data in enumerate(datas):
        out.write_samples(np.array([10]), index=index)
    ths = out.get_reader()
    assert np.array_equal(ths[0].samples[:], np.array([10]))


def test_write_meta_length_1(ets_filenames):
    out = ETSWriter(filename, overwrite=True)

    datas = np.random.randint(100, size=(nb_trace, nb_points), dtype=np.uint8)
    for index, data in enumerate(datas):
        out.write_samples(data, index=index)
        out.write_metadata('plaintext', 22, index=index)

    ths = out.get_reader()
    assert np.array_equal(ths[0].plaintext, np.array([22]))


def test_write_meta_string(ets_filenames):
    out = ETSWriter(filename, overwrite=True)

    datas = np.random.randint(100, size=(nb_trace, nb_points), dtype=np.uint8)
    for index, data in enumerate(datas):
        out.write_samples(data, index=index)
        out.write_metadata('plaintext', 'azerty', index=index)

    ths = out.get_reader()
    assert ths[0].plaintext == 'azerty'


def test_write_samples_1d(ets_filenames):
    out = ETSWriter(filename, overwrite=True)
    datas = np.random.randint(100, size=(nb_trace, nb_points), dtype=np.uint8)
    ths = estraces.formats.read_ths_from_ram(datas)
    out.add_samples(ths[0].samples)
    out.add_samples(ths[1].samples)
    ths_2 = out.get_reader()
    assert np.array_equal(ths[0].samples.array, ths_2[0].samples.array)
    assert np.array_equal(ths[1].samples.array, ths_2[1].samples.array)
    out.close()
    out = ETSWriter(filename)
    out.add_samples(ths[2].samples)
    ths_2 = out.get_reader()
    assert np.array_equal(ths[2].samples.array, ths_2[2].samples.array)


def test_add_samples_2d(ets_filenames):
    out = ETSWriter(filename, overwrite=True)
    datas = np.random.randint(100, size=(nb_trace, nb_points), dtype=np.uint8)
    ths = estraces.formats.read_ths_from_ram(datas)
    out.add_samples(ths.samples)
    ths_2 = out.get_reader()
    assert np.array_equal(ths.samples.array, ths_2.samples.array)


def test_add_samples_raise_exception_if_not_samples_instance(ets_filenames):
    out = ETSWriter(filename)
    with pytest.raises(TypeError):
        out.add_samples([1, 2, 3])


def test_add_samples_truncate_samples_to_first_inserted_data_size(ets_filenames):
    out = ETSWriter(filename, overwrite=True)
    datas = np.random.randint(100, size=(nb_trace, nb_points), dtype=np.uint8)
    ths = estraces.formats.read_ths_from_ram(datas)
    out.add_samples(ths.samples)
    ths_2 = estraces.formats.read_ths_from_ram(np.random.randint(100, size=(nb_trace, nb_points + 10), dtype=np.uint8))
    out.add_samples(ths_2.samples)
    ths_4 = estraces.formats.read_ths_from_ram(np.random.randint(100, size=(nb_trace, nb_points - 10), dtype=np.uint8))
    out.add_samples(ths_4.samples)
    ths_3 = out.get_reader()
    assert len(ths_3[0].samples) == nb_points


def test_writer_open_a_new_file(ets_filenames):
    out = ETSWriter(filename, overwrite=True)
    datas = np.random.randint(100, size=(nb_trace, nb_points), dtype=np.uint8)
    ths = estraces.formats.read_ths_from_ram(datas)
    out.add_samples(ths[0].samples)
    out.close()
    ths_2 = estraces.read_ths_from_ets_file(filename)
    assert np.array_equal(ths_2[0].samples.array, ths[0].samples.array)
    assert len(ths_2) == 1
    os.remove(filename)


def test_writer_open_an_existing_file_and_append_to(ets_filenames):
    out = ETSWriter(filename)
    datas = np.random.randint(100, size=(nb_trace, nb_points), dtype=np.uint8)
    ths = estraces.formats.read_ths_from_ram(datas)
    out.add_samples(ths[0].samples)
    out.close()
    out = ETSWriter(filename)
    out.add_samples(ths[1].samples)
    out.close()
    ths_2 = estraces.read_ths_from_ets_file(filename)
    assert len(ths_2) == 2
    assert np.array_equal(ths_2.samples.array, ths[0:2].samples.array)
    os.remove(filename)


def test_writer_overwrite_an_existing_file(ets_filenames):
    out = ETSWriter(filename)
    datas = np.random.randint(100, size=(nb_trace, nb_points), dtype=np.uint8)
    ths = estraces.formats.read_ths_from_ram(datas)
    out.add_samples(ths[0].samples)
    out.close()
    out = ETSWriter(filename, overwrite=True)
    out.add_samples(ths[1].samples)
    out.close()
    ths_2 = estraces.read_ths_from_ets_file(filename)
    assert len(ths_2) == 1
    assert np.array_equal(ths_2.samples.array, ths[1:2].samples.array)
    os.remove(filename)


def test_ets_writer_add_metadata_for_1_trace(ets_filenames):
    out = ETSWriter(filename)
    datas = np.random.randint(0, 256, size=(nb_trace, nb_points), dtype=np.uint8)
    plaintext = np.random.randint(0, 256, (nb_trace, 16), dtype='uint8')
    chair = np.array(['abcd' for i in range(nb_trace)])
    ciphertext = np.random.randint(0, 256, (nb_trace, 16), dtype='uint8')

    ths = estraces.formats.read_ths_from_ram(
        datas,
        chair=chair,
        plaintext=plaintext,
        ciphertext=ciphertext
    )

    out.add_samples(ths.samples)
    for i in range(nb_trace):
        out.add_metadata(ths[i].metadatas)
    out_ths = out.get_reader()
    assert np.array_equal(out_ths.plaintext, plaintext)
    assert np.array_equal(out_ths.ciphertext, ciphertext)
    assert np.array_equal(out_ths.chair, chair)


def test_ets_writer_add_metadata_for_several_traces(ets_filenames):
    out = ETSWriter(filename)
    datas = np.random.randint(0, 256, size=(nb_trace, nb_points), dtype=np.uint8)
    plaintext = np.random.randint(0, 256, (nb_trace, 16), dtype='uint8')
    chair = np.array(['abcd' for i in range(nb_trace)])
    ciphertext = np.random.randint(0, 256, (nb_trace, 16), dtype='uint8')

    ths = estraces.formats.read_ths_from_ram(
        datas,
        chair=chair,
        plaintext=plaintext,
        ciphertext=ciphertext
    )

    out.add_samples(ths.samples)
    out.add_metadata(ths.metadatas)
    out_ths = out.get_reader()
    assert np.array_equal(out_ths.plaintext, plaintext)
    assert np.array_equal(out_ths.ciphertext, ciphertext)
    assert np.array_equal(out_ths.chair.tolist(), chair.tolist())


def test_ets_writer_raises_exception_if_metadatas_is_not_proper_type(ets_filenames):
    out = ETSWriter(filename)
    with pytest.raises(TypeError):
        out.add_metadata({"dic": 1})


def test_ets_writer_add_metadata_with_inconsistent_sizes(ets_filenames):
    out = ETSWriter(filename)
    datas = np.random.randint(0, 256, size=(nb_trace, nb_points), dtype=np.uint8)
    plaintext = np.random.randint(0, 256, (nb_trace, 16), dtype='uint8')
    chair = np.array(['abcd' for i in range(nb_trace)])
    ciphertext = np.random.randint(0, 256, (nb_trace, 16), dtype='uint8')

    ths = estraces.formats.read_ths_from_ram(
        datas,
        chair=chair,
        plaintext=plaintext,
        ciphertext=ciphertext
    )

    out.add_samples(ths.samples)
    out.add_metadata(ths.metadatas)

    datas = np.random.randint(0, 256, size=(nb_trace, nb_points), dtype=np.uint8)
    plaintext = np.random.randint(0, 256, (nb_trace, 18), dtype='uint8')
    ths = estraces.formats.read_ths_from_ram(
        datas,
        chair=chair,
        plaintext=plaintext,
        ciphertext=ciphertext
    )
    out.add_metadata(ths.metadatas)
    ths = out.get_reader()
    assert ths.plaintext.shape == (2 * nb_trace, 16)


def test_ets_writer_add_trace(ets_filenames):

    out = ETSWriter(filename)
    datas = np.random.randint(0, 256, size=(nb_trace, nb_points), dtype=np.uint8)
    plaintext = np.random.randint(0, 256, (nb_trace, 16), dtype='uint8')
    chair = np.array(['abcd' for i in range(nb_trace)])
    ciphertext = np.random.randint(0, 256, (nb_trace, 16), dtype='uint8')

    ths = estraces.formats.read_ths_from_ram(
        datas,
        chair=chair,
        plaintext=plaintext,
        ciphertext=ciphertext
    )
    out.add_trace(ths[0])

    out_ths = out.get_reader()
    assert np.array_equal(out_ths[0].plaintext, plaintext[0])
    assert np.array_equal(out_ths[0].ciphertext, ciphertext[0])
    assert out_ths[0].chair == chair.tolist()[0]
    assert np.array_equal(out_ths[0].samples.array, datas[0])


def test_ets_writer_add_trace_raises_exception_if_trace_has_improper_types(ets_filenames):
    out = ETSWriter(filename)
    with pytest.raises(TypeError):
        out.add_trace({'not': 'a trace'})


def test_ets_writer_add_trace_raises_exception_if_metadata_are_inconsistent(ets_filenames):
    out = ETSWriter(filename)
    datas = np.random.randint(0, 256, size=(nb_trace, nb_points), dtype=np.uint8)
    plaintext = np.random.randint(0, 256, (nb_trace, 16), dtype='uint8')
    chair = np.array(['abcd' for i in range(nb_trace)])
    ciphertext = np.random.randint(0, 256, (nb_trace, 16), dtype='uint8')

    ths = estraces.formats.read_ths_from_ram(
        datas,
        chair=chair,
        plaintext=plaintext,
        ciphertext=ciphertext
    )
    out.add_trace_header_set(ths)

    datas = np.random.randint(0, 256, size=(nb_trace, nb_points), dtype=np.uint8)
    plaintext_2 = np.random.randint(0, 256, (nb_trace, 16), dtype='uint8')
    ciphertext = np.random.randint(0, 256, (nb_trace, 16), dtype='uint8')

    ths = estraces.formats.read_ths_from_ram(
        datas,
        plaintext_2=plaintext_2,
        ciphertext=ciphertext
    )
    with pytest.raises(estraces.ETSWriterError):
        out.add_trace(ths[0])


def test_ets_writer_add_trace_header_set(ets_filenames):

    out = ETSWriter(filename)
    datas = np.random.randint(0, 256, size=(nb_trace, nb_points), dtype=np.uint8)
    plaintext = np.random.randint(0, 256, (nb_trace, 16), dtype='uint8')
    chair = np.array(['abcd' for i in range(nb_trace)])
    ciphertext = np.random.randint(0, 256, (nb_trace, 16), dtype='uint8')

    ths = estraces.formats.read_ths_from_ram(
        datas,
        chair=chair,
        plaintext=plaintext,
        ciphertext=ciphertext
    )
    new_meta = np.random.randint(0, 255, (len(ths), 1), dtype='uint8')
    ths.metadatas['new'] = new_meta
    out.add_trace_header_set(ths)
    out_ths = out.get_reader()
    assert np.array_equal(out_ths.plaintext, plaintext)
    assert np.array_equal(out_ths.ciphertext, ciphertext)
    assert out_ths.chair.tolist() == chair.tolist()
    assert np.array_equal(out_ths.samples.array, datas)
    assert np.array_equal(out_ths.new, new_meta)


def test_ets_writer_add_ths_raises_exception_if_ths_has_improper_types(ets_filenames):
    out = ETSWriter(filename)
    with pytest.raises(TypeError):
        out.add_trace_header_set({'not': 'a ths'})


def test_ets_writer_add_ths_raises_exception_if_metadata_are_inconsistent(ets_filenames):
    out = ETSWriter(filename)
    datas = np.random.randint(0, 256, size=(nb_trace, nb_points), dtype=np.uint8)
    plaintext = np.random.randint(0, 256, (nb_trace, 16), dtype='uint8')
    chair = np.array(['abcd' for i in range(nb_trace)])
    ciphertext = np.random.randint(0, 256, (nb_trace, 16), dtype='uint8')

    ths = estraces.formats.read_ths_from_ram(
        datas,
        chair=chair,
        plaintext=plaintext,
        ciphertext=ciphertext
    )
    out.add_trace_header_set(ths)

    datas = np.random.randint(0, 256, size=(nb_trace, nb_points), dtype=np.uint8)
    plaintext_2 = np.random.randint(0, 256, (nb_trace, 16), dtype='uint8')
    ciphertext = np.random.randint(0, 256, (nb_trace, 16), dtype='uint8')

    ths = estraces.formats.read_ths_from_ram(
        datas,
        plaintext_2=plaintext_2,
        ciphertext=ciphertext
    )
    with pytest.raises(estraces.ETSWriterError):
        out.add_trace_header_set(ths)


def test_write_trace_with_new_metadata(ets_filenames):
    datas = np.random.randint(0, 256, size=(nb_trace, nb_points), dtype=np.uint8)
    plaintext = np.random.randint(0, 256, (nb_trace, 16), dtype='uint8')
    ciphertext = np.random.randint(0, 256, (nb_trace, 16), dtype='uint8')

    ths = estraces.formats.read_ths_from_ram(
        datas,
        plaintext=plaintext,
        ciphertext=ciphertext
    )

    for index, trace in enumerate(ths):
        trace.new_meta = 'oups'
        out = ETSWriter(filename)
        out.write_trace_object_and_points(trace, trace.samples[:], index)
    out.close()

    ths_2 = estraces.read_ths_from_ets_file(filename)

    assert ths_2[0].new_meta == 'oups'


def test_write_samples(ets_filenames):
    out = ETSWriter(filename, overwrite=True)

    datas = np.random.randint(100, size=(nb_trace, nb_points), dtype=np.uint8)
    plaintext = np.random.randint(256, size=(nb_trace, 16), dtype=np.uint8)
    for index, data in enumerate(datas):
        out.write_samples(data, index=index)
        out.write_metadata('plaintext', plaintext[index], index=index)

    ths = out.get_reader()
    for i, t in enumerate(ths):
        assert np.array_equal(t.samples[:], datas[i])
        assert np.array_equal(t.plaintext, plaintext[i])


def test_ets_writer_accepts_paths_from_pathlib(ets_filenames):
    out = ETSWriter(Path(filename), overwrite=True)

    datas = np.random.randint(100, size=(nb_trace, nb_points), dtype=np.uint8)
    plaintext = np.random.randint(256, size=(nb_trace, 16), dtype=np.uint8)
    for index, data in enumerate(datas):
        out.write_samples(data, index=index)
        out.write_metadata('plaintext', plaintext[index], index=index)
    out.get_reader()


def test_write_trace_with_scalar_metadata(ets_filenames):
    datas = np.random.randint(0, 256, size=(nb_trace, nb_points), dtype=np.uint8)
    scals = np.random.randint(0, 256, (nb_trace,), dtype='uint8')

    ths = estraces.formats.read_ths_from_ram(
        datas,
        scals=scals
    )

    out = ETSWriter(filename)
    out.add_trace_header_set(ths)

    ths_2 = out.get_reader()

    assert len(ths_2.scals) == nb_trace


def test_write_headers(ets_filenames):
    ets = estraces.ETSWriter(filename, overwrite=True)
    ets.write_samples(np.random.randint(0, 255, (1000, 100), dtype='uint8'))
    ets.write_headers(_HEADERS)
    ths = ets.get_reader()
    assert np.array_equal(ths.headers['key'], _HEADERS['key'])
    assert ths.headers['foo'] == _HEADERS['foo']
    assert ths.headers['time'] == _HEADERS['time']


def test_add_trace_header_set_with_headers(ets_filenames):
    ets = estraces.ETSWriter(filename, overwrite=True)
    ths = estraces.read_ths_from_ram(
        np.random.randint(-55000, 55000, (nb_trace, nb_points), dtype='int32'),
        headers=_HEADERS,
        plaintext=np.random.randint(0, 256, (nb_trace, 16), dtype='uint8'),
        ciphertext=np.random.randint(0, 256, (nb_trace, 16), dtype='uint8'),
        chairs=np.array([f'chair{i}' for i in range(nb_trace)])
    )
    ets.add_trace_header_set(ths)
    ths = ets.get_reader()
    assert np.array_equal(ths.headers['key'], _HEADERS['key'])
    assert ths.headers['foo'] == _HEADERS['foo']
    assert ths.headers['time'] == _HEADERS['time']

    ets = estraces.ETSWriter(filename_1, overwrite=True)
    ets.add_trace_header_set(ths)
    ets.write_headers({'bar': 'bar'})
    ths = ets.get_reader()
    assert np.array_equal(ths.headers['key'], _HEADERS['key'])
    assert ths.headers['foo'] == _HEADERS['foo']
    assert ths.headers['time'] == _HEADERS['time']
    assert ths.headers['bar'] == 'bar'


########################################################################################
#  Compression                                                                         #
########################################################################################

@pytest.fixture()
def new_ets(ets_filenames):
    ets = estraces.ETSWriter(filename, overwrite=True)
    ths = estraces.read_ths_from_ram(
        np.random.randint(-55000, 55000, (nb_trace, nb_points), dtype='int32'),
        plaintext=np.random.randint(0, 256, (nb_trace, 16), dtype='uint8'),
        ciphertext=np.random.randint(0, 256, (nb_trace, 16), dtype='uint8'),
        chairs=np.array([f'chair{i}' for i in range(nb_trace)])
    )
    ets.add_trace_header_set(ths)
    ets.write_headers({'foo': 'bar', 'plop': 42})
    ths = ets.get_reader()
    return ths


def test_compress_ets_raises_exception_if_filename_not_provided(ets_filenames):
    with pytest.raises(TypeError):
        estraces.compress_ets(out_filename='foo')
    with pytest.raises(TypeError):
        estraces.compress_ets(filename='foo')


def test_compress_ets_raises_exception_on_bad_types(ets_filenames):
    with pytest.raises(TypeError, match='filename and out_filename must be str or Path, not'):
        estraces.compress_ets(filename=1233, out_filename='foo')
    with pytest.raises(TypeError, match='filename and out_filename must be str or Path, not'):
        estraces.compress_ets(filename='foo', out_filename=123)


def test_compress_ets_raises_exception_if_original_file_doesnt_exist(ets_filenames):
    with pytest.raises(AttributeError):
        estraces.compress_ets(filename='foo', out_filename='bar')


def test_compress_ets_raises_exception_if_original_outfile_exist(new_ets):
    with pytest.raises(ValueError):
        estraces.compress_ets(filename='foo', out_filename=filename)
    with pytest.raises(ValueError):
        estraces.compress_ets(filename=filename, out_filename=filename)


def test_compress_ets(new_ets):
    estraces.compress_ets(filename=filename, out_filename=compressed)
    ths_comp = estraces.read_ths_from_ets_file(compressed)
    assert np.array_equal(new_ets.samples[:], ths_comp.samples[:])
    assert np.array_equal(new_ets.plaintext, ths_comp.plaintext)
    assert np.array_equal(new_ets.ciphertext, ths_comp.ciphertext)
    assert np.array_equal(new_ets.chairs, ths_comp.chairs)
    assert os.path.getsize(filename) > 1.1 * os.path.getsize(compressed)


def test_compress_ets_accepts_paths_from_pathlib(new_ets):
    estraces.compress_ets(filename=Path(filename), out_filename=Path(compressed))
    estraces.read_ths_from_ets_file(Path(compressed))


def test_compress_ets_keep_headers(new_ets):
    estraces.compress_ets(filename=filename, out_filename=compressed)
    ths_comp = estraces.read_ths_from_ets_file(compressed)
    assert len(dict(ths_comp.headers)) > 0
    assert dict(ths_comp.headers) == dict(new_ets.headers)


def _get_virtual_memory_mock(v):
    def virtual_memory_wrapped():
        return 1, v
    return virtual_memory_wrapped


@patch("psutil.virtual_memory", side_effect=_get_virtual_memory_mock(20_000_000))
def test_get_optimal_batch_size_with_size_lower_available_memory(virtual_memory_func):
    ths = estraces.read_ths_from_ram(
        np.random.randint(-55000, 55000, (nb_trace, nb_points), dtype='int32'),
        headers=_HEADERS,
        plaintext=np.random.randint(0, 256, (nb_trace, 16), dtype='uint8'),
        ciphertext=np.random.randint(0, 256, (nb_trace, 16), dtype='uint8'),
        chairs=np.array([f'chair{i}' for i in range(nb_trace)])
    )
    batch_size = _get_optimal_batch_size(ths)
    assert batch_size > 1


@patch("psutil.virtual_memory", side_effect=_get_virtual_memory_mock(1))
def test_get_optimal_batch_size_with_size_higher_available_memory(virtual_memory_func):
    ths = estraces.read_ths_from_ram(
        np.random.randint(-55000, 55000, (nb_trace, nb_points), dtype='int32'),
        headers=_HEADERS,
        plaintext=np.random.randint(0, 256, (nb_trace, 16), dtype='uint8'),
        ciphertext=np.random.randint(0, 256, (nb_trace, 16), dtype='uint8'),
        chairs=np.array([f'chair{i}' for i in range(nb_trace)])
    )
    with pytest.raises(MemoryError):
        _get_optimal_batch_size(ths)


@patch("psutil.virtual_memory", side_effect=_get_virtual_memory_mock(32e9))
def test_get_optimal_batch_size_takes_limit_into_account(virtual_memory_func):
    ths = estraces.read_ths_from_ram(
        np.random.randint(-55000, 55000, (nb_trace, nb_points), dtype='int32'),
        headers=_HEADERS,
        plaintext=np.random.randint(0, 256, (nb_trace, 16), dtype='uint8'),
        ciphertext=np.random.randint(0, 256, (nb_trace, 16), dtype='uint8'),
        chairs=np.array([f'chair{i}' for i in range(nb_trace)])
    )
    reference_batch_size = _get_optimal_batch_size(ths, available_memory_ratio=1, limit=256e9)
    assert _get_optimal_batch_size(ths, available_memory_ratio=1, limit=16e9) == reference_batch_size // 2
    assert _get_optimal_batch_size(ths, available_memory_ratio=1, limit=17e9) != reference_batch_size // 2



from ..traces.abstract_reader import AbstractReader
from ..traces.trace_header_set import build_trace_header_set
import os as _os
import numpy as _np
import typing
import sqlite3
from typing import NamedTuple, Dict


class SQLiteError(Exception):
    pass


class SamplesConfig(NamedTuple):
    table_name: str = 'traces'
    pk_column_name: str = 'trace_id'
    samples_column_name: str = 'samples'
    dtype: str = 'uint8'


class MetadataConfig(NamedTuple):
    metadata_defs: Dict = {}
    table_name: str = 'traces'
    pk_column_name: str = 'trace_id'


class HeadersConfig(NamedTuple):
    table_name: str = 'headers'
    pk_column_name: str = 'id'
    tag_column_name: str = 'tag'
    value_column_name: str = 'value'
    dtype_column_name: str = 'dtype'


EmptyMetadata = MetadataConfig(table_name=None)
EmptyHeaders = HeadersConfig(table_name=None)


def read_ths_from_sqlite(
    filename,
    samples_config: SamplesConfig,
    metadata_config: MetadataConfig = EmptyMetadata,
    headers_config: HeadersConfig = EmptyHeaders
):
    """Build and return a :class:`TraceHeaderSet` instance read from a SQLite3 database file.

    Args:
        filename (str or Path): filename of the SQLite3 database.
        samples_config (SamplesConfig): instance of `SamplesConfig` with parameters to read samples - table name, pk column name, samples column name and dtype
            format of samples data.
        metadata_config (MetadataConfig): instance of `MetadataConfig` with parameters to read trace metadata - table name, pk column name and dict where each
            entry is the column name of a metadata and the corresponding value the dtype of the metadata.
        headers_config (HeadersConfig): instance of `HeadersConfig` with parameters to read headers from a specific table - table name, tag column name,
            value column name and dtype column name.

    Returns:
        (:obj:`TraceHeaderSet`)

    """
    return build_trace_header_set(
        reader=SQLiteFormatReader(
            filename,
            samples_config=samples_config,
            metadata_config=metadata_config,
            headers_config=headers_config
        ),
        name='SQLite reader'
    )


class SQLiteFormatReader(AbstractReader):

    def __init__(self, filename: str, samples_config: SamplesConfig, metadata_config: MetadataConfig, headers_config: HeadersConfig):
        if not _os.path.exists(filename):
            raise ValueError(f'SQLite database file {filename} does not exists.')
        self._connection = sqlite3.connect(filename)
        try:
            self._connection.cursor().execute('select name from sqlite_master')
        except sqlite3.DatabaseError:
            raise ValueError(f'{filename} is not a SQLite3 database.')
        self._filename = filename
        self._samples_config = samples_config
        self._metadata_config = metadata_config
        self._headers_config = headers_config
        self.dtype = samples_config.dtype
        self._sub_traceset_indices = None
        self._headers_keys = None

    def __len__(self):
        if self._size is None:
            query = f'SELECT {self._samples_config.pk_column_name} FROM {self._samples_config.table_name}'
            data = _execute_query(self._connection, query, (), error_message='Problem with samples configuration: ')
            self._size = len(data)
        return self._size

    def fetch_samples(self, traces: list, frame=None) -> _np.ndarray:
        """Fetch samples for the given traces id and given samples data frame.

        Args:
            traces: Lists of traces id to fetch.
            frame: Samples data to fetch. Must support `Ellipsis`, `slice`, `list`, `ndarray` or `int` types.

        Returns:
            (:class:`numpy.ndarray`) array of shape (number of traces, size of samples)

        """
        if isinstance(traces, range) and (traces.start < 0 or traces.stop < 0):
            traces = range(self._size + traces.start, self._size + traces.stop, traces.step)

        traces = [i if i >= 0 else self._size + i for i in traces]
        traces = self._convert_traces_indices_to_db_indices(traces)

        ordered_unique_ids = sorted(set(traces))
        mask_needed = ordered_unique_ids != traces.tolist()

        ids = ','.join([str(t) for t in traces])

        query = f'''
        SELECT {self._samples_config.pk_column_name}, {self._samples_config.samples_column_name}
        FROM {self._samples_config.table_name} WHERE {self._samples_config.pk_column_name} in ({ids})
        '''
        data = _execute_query(self._connection, query, (), error_message='Error with samples configuration: ')
        res = _np.array([_np.frombuffer(d[1], dtype=self.dtype) for d in data])

        if mask_needed:
            mask = _np.zeros((len(traces), len(res)), dtype='uint8')
            for i in range(len(traces)):
                mask[i, ordered_unique_ids.index(traces[i])] = 1
            res = mask.dot(res)

        if isinstance(frame, int):
            frame = slice(frame, frame + 1) if frame >= 0 else slice(res[0].shape[0] + frame, res[0].shape[0] + 1 + frame)
        if res.ndim > 1:
            res = res[:, frame]
        return res

    def fetch_metadatas(self, key: typing.Hashable, trace_id: int = None):
        """Fetch metadata value for the given metadata key and trace id.

        Args:
            key (typing.Hashable): Key of the metadata to fetch. Must be hashable.
            trace_id (int): Trace id for which to fetch the metadata.

        Returns:
            A container of all the values of the trace set for the given metadata if trace_id is None.
            Else, the value of the metadata for the given trace id.

        """
        if self._metadata_config is not EmptyMetadata:
            if trace_id is not None:
                trace_id = self._convert_trace_index_to_db_index(trace_id)
                query = f'SELECT {key} from {self._metadata_config.table_name} where {self._metadata_config.pk_column_name}={trace_id}'
                data = _execute_query(self._connection, query, (), error_message=f'Error with metadata {key}: ')[0][0]
                if isinstance(data, str):
                    return data
                return _np.frombuffer(data, dtype=self._metadata_config.metadata_defs[key])
            else:
                query = f'SELECT {key} FROM {self._metadata_config.table_name}'
                data = _execute_query(self._connection, query, (), error_message=f'Error with metadata {key}: ')
                res = []
                for t in data:
                    if isinstance(t[0], str):
                        res.append(t[0])
                    else:
                        res.append(_np.frombuffer(t[0], dtype=self._metadata_config.metadata_defs[key]))
                return _np.array(res)[self._traceset_indices]

    @property
    def _traceset_indices(self):
        return self._sub_traceset_indices if self._sub_traceset_indices is not None else slice(None, None, 1)

    def _convert_trace_index_to_db_index(self, trace_id):
        return self._convert_traces_indices_to_db_indices([trace_id])[0]

    def _convert_traces_indices_to_db_indices(self, traces):
        if self._sub_traceset_indices is not None:
            sub_max = len(self._sub_traceset_indices)
            traces_index = [t for t in traces if t < sub_max]
            return self._sub_traceset_indices[traces_index]
        return _np.array(traces)

    def __getitem__(self, key):
        """Return a new format instance limited to traces[key] subset.

        Args:
            key: slice or list of traces indexes to slice on.

        """
        super().__getitem__(key)
        if isinstance(key, int):
            key = [key]
        elif isinstance(key, slice):
            key = range(
                key.start if key.start is not None else 0,
                key.stop if key.stop is not None else len(self),
                key.step if key.step is not None else 1
            )
        new_reader = SQLiteFormatReader(
            filename=self._filename,
            metadata_config=self._metadata_config,
            samples_config=self._samples_config,
            headers_config=self._headers_config
        )
        sub_traceset_indices = self._convert_traces_indices_to_db_indices(traces=key)
        traces_number = len(sub_traceset_indices)
        new_reader._sub_traceset_indices = sub_traceset_indices
        new_reader._size = traces_number
        return new_reader

    def fetch_header(self, key: typing.Hashable):
        """Fetch header value for the given key.

        Args:
            key (typing.Hashable): key of the header to fetch.

        Returns:
            the header value.

        """
        if self._headers_config is not EmptyHeaders:
            query = f'''
            SELECT {self._headers_config.value_column_name}, {self._headers_config.dtype_column_name}
            FROM {self._headers_config.table_name} where {self._headers_config.tag_column_name}="{key}"
            '''
            data = _execute_query(self._connection, query, (), error_message='Problem with headers config: ')
            if isinstance(data[0][0], str):
                return data[0][0]
            else:
                return _np.frombuffer(data[0][0], dtype=data[0][1])

    @property
    def metadatas_keys(self):
        """Provide a list or views of the metadatas keys available."""
        if self._metadata_config is not EmptyMetadata:
            return self._metadata_config.metadata_defs.keys()
        return []

    def get_trace_size(self, trace_id):
        """Provide the size of trace trace_id."""
        query = f'SELECT {self._samples_config.samples_column_name} FROM {self._samples_config.table_name} WHERE {self._samples_config.pk_column_name}=?'
        data = _execute_query(self._connection, query, (trace_id, ), error_message='Problem with samples configuration: ')
        data = _np.frombuffer(data[0][0], dtype=self.dtype)
        return data.shape[0]

    @property
    def headers_keys(self):
        """Provide a list or view of the headers keys available."""
        if self._headers_keys is None and self._headers_config is not EmptyHeaders:
            query = f'SELECT {self._headers_config.tag_column_name} FROM {self._headers_config.table_name}'
            data = _execute_query(self._connection, query, (), 'Problem with headers config: ')
            self._headers_keys = [t[0] for t in data]
        elif self._headers_keys is None:
            self._headers_keys = []
        return self._headers_keys


def _execute_query(connection, query, query_args, error_message='Problem executing query:'):
    try:
        return connection.cursor().execute(query, query_args).fetchall()
    except sqlite3.OperationalError as e:
        raise SQLiteError(f'{error_message} {e}.')



from .context import estraces # noqa
import pytest
from estraces.formats import concat_format, ets_format
import numpy as np
import cloudpickle


path = 'tests/samples/test.ets'


@pytest.fixture
def ets1():
    return ets_format.read_ths_from_ets_file(filename=path)


@pytest.fixture
def concat1():
    ets1 = ets_format.read_ths_from_ets_file(filename=path)
    ets2 = ets_format.read_ths_from_ets_file(filename=path)
    return concat_format.read_ths_from_multiple_ths(ets1, ets2)


def test_instanciate(concat1):
    assert isinstance(concat1, estraces.TraceHeaderSet)
    assert 400 == len(concat1)


def test_sub_ths(concat1):
    concat2 = concat1[:40]
    assert isinstance(concat2, estraces.TraceHeaderSet)
    assert 40 == len(concat2)


def test_raises_type_error_if_args_are_not_all_traceheadersets(ets1):
    with pytest.raises(TypeError):
        concat_format.read_ths_from_multiple_ths(ets1, "foo_bar")


def test_split_preserves_length_consistency_of_sub_ths(concat1):
    parts = concat1.split(100)
    for _, p in enumerate(parts):
        assert len(p) == 100
        assert len(p) == p.samples[:].shape[0]


inconsistent_trace_sets = ['test_without_plaintext.ets',
                           'test_with_plaintext_too_short.ets',
                           'test_with_plaintext_wrong_dtype.ets',
                           'test_with_plaintext_wrong_name.ets',
                           'test_with_samples_too_short.ets',
                           'test_with_samples_wrong_dtype.ets']


@pytest.fixture(params=inconsistent_trace_sets)
def ets2(request):
    path = 'tests/samples/' + request.param
    return ets_format.read_ths_from_ets_file(filename=path)


def test_wrong_meta_and_samples_consistency_raises(ets1, ets2):
    with pytest.raises(ValueError):
        concat_format.read_ths_from_multiple_ths(ets1, ets2)


def test_meta_strings_with_different_lengths():
    ets1 = ets_format.read_ths_from_ets_file(filename='tests/samples/test_str_meta_1.ets')
    ets2 = ets_format.read_ths_from_ets_file(filename='tests/samples/test_str_meta_2.ets')
    ets = concat_format.read_ths_from_multiple_ths(ets1, ets2)

    str_stack = np.concatenate((ets1.str_meta[:], ets2.str_meta[:]), axis=0)
    np_str_stack = np.concatenate((ets1.np_str_meta[:], ets2.np_str_meta[:]), axis=0)

    assert np.array_equal(ets.str_meta[:], str_stack)
    assert np.array_equal(ets.np_str_meta[:], np_str_stack)


def test_inconsistent_headers_raises_exception():
    ths_1 = estraces.read_ths_from_ram(
        np.random.randint(0, 255, (1000, 1200), dtype='uint8'),
        headers={'foo': 'ths_1', 'bar': 'bar'},
        plaintext=np.random.randint(0, 255, (1000, 16), dtype='uint8')
    )
    ths_2 = estraces.read_ths_from_ram(
        np.random.randint(0, 255, (1000, 1200), dtype='uint8'),
        headers={'foo': 'ths_2', 'bar': 'bar'},
        plaintext=np.random.randint(0, 255, (1000, 16), dtype='uint8')
    )
    with pytest.raises(ValueError):
        concat_format.read_ths_from_multiple_ths(ths_1, ths_2)
    ths_1 = estraces.read_ths_from_ram(
        np.random.randint(0, 255, (1000, 1200), dtype='uint8'),
        headers={'foo': 'ths_1', 'bar': 'bar'},
        plaintext=np.random.randint(0, 255, (1000, 16), dtype='uint8')
    )
    ths_2 = estraces.read_ths_from_ram(
        np.random.randint(0, 255, (1000, 1200), dtype='uint8'),
        headers={'foo_2': 'ths_2', 'bar': 'bar'},
        plaintext=np.random.randint(0, 255, (1000, 16), dtype='uint8')
    )
    with pytest.raises(ValueError):
        concat_format.read_ths_from_multiple_ths(ths_1, ths_2)


def test_headers_can_be_overridden():
    ths_1 = estraces.read_ths_from_ram(
        np.random.randint(0, 255, (1000, 1200), dtype='uint8'),
        headers={'foo': 'ths_1', 'bar': 'bar', 'spec': 1},
        plaintext=np.random.randint(0, 255, (1000, 16), dtype='uint8')
    )
    ths_2 = estraces.read_ths_from_ram(
        np.random.randint(0, 255, (1000, 1200), dtype='uint8'),
        headers={'foo': 'ths_2', 'bar': 'bar', 'true': 2},
        plaintext=np.random.randint(0, 255, (1000, 16), dtype='uint8')
    )
    with pytest.raises(ValueError):
        concat_format.read_ths_from_multiple_ths(ths_1, ths_2)

    ths = concat_format.read_ths_from_multiple_ths(
        ths_1, ths_2,
        headers={
            'foo': 'concat ths',
            'spec': 1,
            'true': 2,
            'new_header': 'new'
        }
    )
    assert sorted(list(ths.headers.keys())) == ['bar', 'foo', 'new_header', 'spec', 'true']
    assert ths.headers['foo'] == 'concat ths'
    assert ths.headers['bar'] == 'bar'
    assert ths.headers['spec'] == 1
    assert ths.headers['true'] == 2
    assert ths.headers['new_header'] == 'new'


def test_serialization(concat1):
    ths = concat1
    dumped = cloudpickle.dumps(ths)
    loaded_ths = cloudpickle.loads(dumped)
    assert np.array_equal(loaded_ths.samples[:], ths.samples[:])
    assert np.array_equal(loaded_ths.plaintext[:], ths.plaintext[:])
    assert np.array_equal(loaded_ths.foo_bar[:], ths.foo_bar[:])
    assert np.array_equal(loaded_ths.ciphertext[:], ths.ciphertext[:])


def test_concat_keeps_order():
    ths_list = [estraces.read_ths_from_ram(np.random.randint(0, 255, (100, 120), dtype='uint8')) for i in range(10)]

    ths = concat_format.read_ths_from_multiple_ths(*ths_list)
    for i in range(10):
        assert np.array_equal(ths_list[i].samples[:], ths.samples[i * 100:(i + 1) * 100]), 'Concatenation does not guarantee order.'


# see https://gitlab.com/eshard/estraces/-/issues/45
def test_filename_property(concat1):
    assert concat1.filename is None



from .context import estraces  # noqa
import pytest
import numpy as np
from .conftest import _HEADERS


@pytest.fixture(params=['1 trace', 'all traces'])
def headers(request, fmt):
    return estraces.traces.headers.Headers(reader=fmt)


def test_headers_init_with_improper_reader_type_raises_exception():
    with pytest.raises(TypeError):
        estraces.traces.headers.Headers(reader="oups")


def test_headers_repr(headers):
    assert repr(headers) == str(headers)


def test_read_headers_key_throws_key_error_on_missing_key(headers):
    with pytest.raises(KeyError):
        headers["missing"]
    with pytest.raises(KeyError):
        headers["3"]


def test_read_headers_key_throws_type_error_on_incorrect_key_type(headers):
    with pytest.raises(TypeError):
        headers[np.array(range(2))]
    with pytest.raises(TypeError):
        headers[np.array([1, 2, 3], dtype="uint8")]


def test_read_headers_returns_a_mapping(headers):
    keys = [k for k in sorted(headers.keys())]
    assert 'foo' in keys
    assert 'key' in keys
    assert 'time' in keys
    for k, v in headers.items():
        if k == 'key':
            assert np.array_equal(v, _HEADERS[k])
        elif k == 'foo' or k == 'time':
            assert v == _HEADERS[k]


def test_headers_equality_with_not_mapping_type_raise_exception(headers):
    with pytest.raises(NotImplementedError):
        assert headers == 2


def test_headers_equality(headers, fmt):
    d = dict(headers)
    d_1 = dict(headers)
    d_1.update({'another_key': 122390490})
    assert headers == d
    assert headers != d_1
    h_2 = estraces.traces.headers.Headers(fmt)
    assert headers == h_2
    assert id(headers) != id(h_2)



from .context import estraces  # noqa: F401
from estraces import read_ths_from_trs_file
import pytest
import numpy as np
import cloudpickle
import trsfile


TRS_FILENAME = 'tests/samples/aes128_sb_ciph_0fec9ca47fb2f2fd4df14dcb93aa4967.trs'  # TRS V1 file

PLAINTEXTS_10 = np.array([[111, 111, 0, 202, 89, 18, 30, 159, 41, 95, 134, 24, 243, 193, 171, 27],
                          [212, 44, 106, 221, 40, 124, 241, 26, 4, 129, 240, 57, 201, 59, 120, 84],
                          [225, 198, 49, 234, 172, 146, 40, 214, 209, 83, 29, 53, 226, 63, 44, 201],
                          [31, 20, 136, 25, 124, 69, 70, 19, 253, 3, 130, 45, 143, 134, 119, 153],
                          [102, 189, 40, 64, 172, 191, 65, 96, 107, 114, 53, 0, 68, 122, 195, 106],
                          [38, 184, 241, 197, 204, 242, 97, 144, 177, 186, 87, 227, 10, 135, 74, 128],
                          [80, 42, 250, 44, 133, 32, 174, 44, 185, 160, 95, 20, 58, 186, 216, 105],
                          [186, 88, 211, 113, 164, 119, 106, 12, 116, 102, 82, 27, 32, 90, 6, 44],
                          [141, 194, 214, 51, 137, 21, 44, 34, 190, 116, 10, 153, 179, 135, 195, 191],
                          [105, 120, 11, 202, 136, 103, 43, 18, 215, 112, 151, 64, 92, 234, 159, 113]], dtype='uint8')


@pytest.fixture
def trs():
    return read_ths_from_trs_file(TRS_FILENAME, dtype='float32',
                                  metadatas_parsers=dict(plaintext=lambda x: np.frombuffer(x, dtype='uint8', offset=0, count=16),
                                                         ciphertext=lambda x: np.frombuffer(x, dtype='uint8', offset=16)))


def test_trs_file_raise_exception_if_filename_incorrect():
    with pytest.raises(AttributeError):
        read_ths_from_trs_file(filename='fakefilename', metadatas_parsers={})

    with pytest.raises(AttributeError):
        read_ths_from_trs_file(filename=12334, metadatas_parsers={})

    with pytest.raises(AttributeError):
        read_ths_from_trs_file(filename='tests/samples/test.ets', metadatas_parsers={})


def test_trs_reader_raise_exception_if_metadatas_parser_incorrect():
    with pytest.raises(AttributeError):
        read_ths_from_trs_file(filename=TRS_FILENAME, metadatas_parsers=[1, 2])

    with pytest.raises(AttributeError):
        read_ths_from_trs_file(filename=TRS_FILENAME, metadatas_parsers="ffgfgfgf")

    with pytest.raises(AttributeError):
        read_ths_from_trs_file(filename=TRS_FILENAME, metadatas_parsers={"meta1": "Yipi"})

    with pytest.raises(AttributeError):
        read_ths_from_trs_file(filename=TRS_FILENAME, metadatas_parsers={"meta1": 1})

    with pytest.raises(AttributeError):
        read_ths_from_trs_file(filename=TRS_FILENAME, metadatas_parsers={"meta1": lambda x, y: x + y})

    with pytest.raises(AttributeError):
        def _(x, y):
            x + y
        read_ths_from_trs_file(filename=TRS_FILENAME, metadatas_parsers={"meta1": _})

    with pytest.raises(AttributeError):
        def _(x, y):
            x + y
        read_ths_from_trs_file(filename=TRS_FILENAME, metadatas_parsers={"meta1": lambda x: "constant", "meta2": _})


def test_trs_file_raises_exception_if_dtype_is_incorrect():
    with pytest.raises(TypeError):
        read_ths_from_trs_file(filename=TRS_FILENAME, metadatas_parsers={}, dtype='toto')

    with pytest.raises(TypeError):
        read_ths_from_trs_file(filename=TRS_FILENAME, metadatas_parsers={}, dtype='uuint8')


def test_trs_reader_optional_dtype_overrides_default():
    ths = read_ths_from_trs_file(filename=TRS_FILENAME, metadatas_parsers={})
    assert ths.samples[0, :].dtype == 'int8'
    ths = read_ths_from_trs_file(filename=TRS_FILENAME, metadatas_parsers={}, dtype='float32')
    assert ths.samples[0, :].dtype == 'float32'
    ths = read_ths_from_trs_file(filename=TRS_FILENAME, metadatas_parsers={}, dtype=np.uint16)
    assert ths.samples[0, :].dtype == 'uint16'
    ths = read_ths_from_trs_file(filename=TRS_FILENAME, metadatas_parsers={}, dtype='uint8')
    assert ths.samples[0, :].dtype == 'uint8'


def test_trs_headers_provides_native_file_format_headers():
    ths = read_ths_from_trs_file(filename=TRS_FILENAME, metadatas_parsers={})
    assert dict(ths.headers) == {'title_space': 0, 'sample_coding': 1, 'length_data': 32, 'number_samples': 1920, 'number_traces': 500, 'trace_block': None}


def test_serialization(trs):
    dumped = cloudpickle.dumps(trs)
    loaded_ths = cloudpickle.loads(dumped)
    assert np.array_equal(loaded_ths.samples[:], trs.samples[:])
    assert np.array_equal(loaded_ths.plaintext[:], trs.plaintext[:])
    assert np.array_equal(loaded_ths.ciphertext[:], trs.ciphertext[:])


def test_simple_indexing(trs):
    assert np.array_equal(trs.samples[:10], trs[:10].samples[:])
    assert np.array_equal(trs[:10].plaintext, PLAINTEXTS_10)
    assert np.array_equal(trs[[1, 3, 5]].plaintext, PLAINTEXTS_10[[1, 3, 5]])


def test_simple_indexing_by_integer(trs):
    trs.samples[3]
    assert np.array_equal(trs[3].plaintext, PLAINTEXTS_10[3])


def test_double_indexing_1(trs):
    sub_trs = trs[:20]
    assert np.array_equal(sub_trs.samples[:10], sub_trs[:10].samples[:])
    assert np.array_equal(sub_trs[:10].plaintext, PLAINTEXTS_10)
    assert np.array_equal(sub_trs[[1, 3, 5]].plaintext, PLAINTEXTS_10[[1, 3, 5]])


def test_double_indexing_2(trs):
    sub_trs = trs[[1, 3, 5, 7, 9, 11]]
    assert np.array_equal(sub_trs[:4].plaintext, PLAINTEXTS_10[[1, 3, 5, 7]])
    assert np.array_equal(sub_trs[[1, 3]].plaintext, PLAINTEXTS_10[[3, 7]])


def test_double_indexing_by_integer(trs):
    # requires an independent test to avoid caching effect
    sub_trs = trs[:20]
    sub_trs.samples[3]
    assert np.array_equal(sub_trs[3].plaintext, PLAINTEXTS_10[3])
    sub_trs = trs[[1, 3, 5, 7, 9, 11]]
    assert np.array_equal(sub_trs[3].plaintext, PLAINTEXTS_10[7])


@pytest.mark.skipif(trsfile.__version__ == '0.3.2', reason="Test only for trsfile v2.")
def test_read_trs_v2_file():
    filename = 'tests/samples/test_without_parser.trs'
    trs = read_ths_from_trs_file(filename, dtype='float32',
                                 metadatas_parsers=dict(plaintext=lambda x: np.frombuffer(x, dtype='uint8', offset=0, count=16),
                                                        ciphertext=lambda x: np.frombuffer(x, dtype='uint8', offset=16, count=16)))
    ref = {'title_space': 255, 'sample_coding': 2, 'length_data': 34, 'number_samples': 33, 'number_traces': 10, 'trace_block': None}
    for key, value in ref.items():
        print(key)
        assert trs.headers[key] == value
    assert trs[:3].samples[:].shape == (3, 33)
    assert trs[:3].ciphertext[:].shape == (3, 16)
    assert trs[:3].plaintext[:].shape == (3, 16)


@pytest.mark.skipif(trsfile.__version__ == '0.3.2', reason="Test only for trsfile v2.")
def test_trs_without_metadata_parser():
    filename = 'tests/samples/test_without_parser.trs'
    trs = read_ths_from_trs_file(filename, dtype='float32', metadatas_parsers=None)
    assert trs[:3].samples[:].shape == (3, 33)
    assert trs[:3].ciphertext[:].shape == (3, 16)
    assert trs[:3].plaintext[:].shape == (3, 16)
    assert trs[:3].indice[:].shape == (3, 1)


@pytest.mark.skipif(trsfile.__version__ == '0.3.2', reason="Test only for trsfile v2.")
def test_trsv1_without_metadata_parser_raises():
    with pytest.raises(TypeError, match='The TRS file does not have a TRACE_PARAMETER_DEFINITIONS field. You must set the'):
        read_ths_from_trs_file(TRS_FILENAME, dtype='float32', metadatas_parsers=None)


@pytest.mark.skipif(trsfile.__version__ != '0.3.2', reason="Test only for trsfile v1.")
def test_old_trsfile_version_without_metadata_parser_raises():
    with pytest.raises(TypeError, match='Setting a TRS reader with `metadatas_parsers=None` allowed only '):
        read_ths_from_trs_file(TRS_FILENAME, dtype='float32', metadatas_parsers=None)


# see https://gitlab.com/eshard/estraces/-/issues/45
def test_filename_property(trs):
    assert trs.filename == TRS_FILENAME



from .context import estraces  # noqa
import pytest
import numpy as np
from .conftest import PLAINS, CIPHERS, KEY, TIME, FOO


@pytest.fixture
def ths(fmt):
    return estraces.traces.trace_header_set.build_trace_header_set(reader=fmt, name="Test THS")


@pytest.fixture
def trace(ths):
    return ths.traces[0]


def test_metadatas_is_metadatas_instance(trace):
    assert isinstance(trace.metadatas, estraces.traces.metadatas.Metadatas)


def test_read_metadatas_with_attribute(trace):
    assert np.array_equal(trace.metadatas["ciphertext"], trace.ciphertext)
    assert np.array_equal(trace.ciphertext, CIPHERS[0])
    assert np.array_equal(trace.metadatas["plaintext"], trace.plaintext)
    assert np.array_equal(trace.plaintext, PLAINS[0])


def test_read_metadatas_attributes_missing_raises_attribute_error(trace):
    with pytest.raises(AttributeError):
        trace.myproperty
    with pytest.raises(AttributeError):
        trace.ciphers


def test_trace_samples_is_samples_type(trace):
    assert isinstance(trace.samples, estraces.traces.samples.Samples)
    assert len(trace.samples) == 1000


def test_attribute_headers_returns_headers_type(trace):
    assert isinstance(trace.headers, estraces.traces.headers.Headers)


def test_attribute_headers(trace):
    assert np.array_equal(trace.headers['key'], KEY)
    assert trace.headers['time'] == TIME
    assert trace.headers['foo'] == FOO


def test_add_metadata_to_trace(trace):
    trace.metadatas['new_meta'] = 'foo'
    assert trace.metadatas['new_meta'] == 'foo'
    assert trace.new_meta == 'foo'


def test_add_metadata_to_trace_through_attribute(trace):
    trace.new_meta = 'foo'
    assert trace.new_meta == 'foo'
    assert trace.metadatas['new_meta'] == 'foo'


def test_add_existing_metadata_or_attribute_raises_exception(trace):
    with pytest.raises(AttributeError):
        trace.plaintext = 'foo'
    with pytest.raises(AttributeError):
        trace.samples = 'foo'
    trace.__doc__ = 'foo'
    assert trace.__doc__ == 'foo'
    with pytest.raises(KeyError):
        trace.metadatas['__doc__']



# -*- coding: utf-8 -*-
from . import trace, samples, metadatas, _metaclass, headers
from . import abstract_reader
from .._legacy_formats.trace import TraceDeprecated
from .._legacy_formats.trace_header_set import TraceHeaderSetDeprecated
from collections import abc
import numpy as _np


class TracesContainer(abc.Collection):

    def __init__(self, trace_klass, reader, sub_slice=None):
        self._trace_klass = trace_klass
        self._reader = reader
        self._traces = {}
        self._sub_slice = sub_slice if sub_slice else slice(0, len(self._reader), 1)
        self._keys_list = range(len(self._reader))[self._sub_slice]

    def __getitem__(self, key):
        if isinstance(key, slice):
            return TracesContainer(trace_klass=self._trace_klass, reader=self._reader, sub_slice=key)
        key = self._keys_list[key]
        if key not in self._traces:
            self._traces[key] = self._trace_klass(trace_id=key, reader=self._reader)
        return self._traces[key]

    def __len__(self):
        return len(self._keys_list)

    def __contains__(self, trace):
        return trace._id in self._keys_list

    def __iter__(self):
        for trc in range(len(self)):
            yield self[trc]


class TraceHeaderSet:
    """Provide a consistent API for manipulating samples data and metadatas from any kind of trace files.

    Attributes:
        name (str): name of traces set
        metadatas (:class:`Metadatas`): dict-like metadatas object
        headers (:class:`Headers`): mapping of headers for this trace set, ie one off value metadata.
        samples (:class:`Samples`): 2 dimensions samples object
        traces (list): list of :class:`Trace` instances

    Notes:
        Each metadata available in `metadatas` attribute can be reached through a property.

    Examples:
        With a :class:`TraceHeaderSet` instance you can manipulate underlying :class:`Traces`, samples data,
        metadatas, headers, slice, iterate or filter on your set.

        Iterate on :class:`TraceHeaderSet` instance::

            for trace in ths:
                trace # each item is a :class:`Trace` instance

        Slice is a new :class:`TraceHeaderSet` instance limited to the slice::

            new_ths = ths[:200]
            new_ths = ths[[1, 2, 100]] # list-based slice supported

        Get a :class:`Trace` instance with an integer::

            trace = ths[200] # get item returns a :class:`Trace` instance

        You can also get the traces iterable::

            ths.traces

        Samples are available as 2 dimensional array-like :class:`Samples` with shape (number of traces, size of traces).
        It supports a subset of numpy.ndarray-like slicing - including advanced list-based slicing::

            ths.samples[2000:3000] # slice on traces, keeping all samples
            ths.samples[:, 1000:2000] # slice on samples data, for all traces
            ths.samples[[1, 100, 1000], :] # get samples for traces with indexes 1, 100 and 100

        Metadatas are available through a dict-like :class:`Metadatas` instance::

            metadatas = ths.metadatas
            metadatas['plaintext']

        Each metadata can be reached with its own attribute::

            ths.plaintext # is equivalent to ths.metadatas['plaintext']

        Headers are metadata which have only one value for all the trace set. It is hold on a mapping, dict like object:

            ths.headers['key']

        Filter traces in your ths with any arbitrary filtering function.
        As an example, get a new TraceHeaderSet with traces which first samples value is even::

            filtered_ths = ths.filter(lambda trace: trace.samples[0] % 2 == 0)

    Warning:
        TraceHeaderSet instance must be obtained through the use of a factory function suited to the use of your files format.

    """

    _is_valid_trace_class = False
    _is_initialized = False

    def __init__(self, reader, name=""):
        """Must not be instantiated or initialized directly, only through :func:`build_trace_header_set`."""
        if not self.__class__._is_valid_trace_class:
            raise TypeError(
                "{c} cant be instantiate directly. Use {f} function.".format(
                    c=self.__class__.__name__, f=build_trace_header_set.__name__
                )
            )
        try:
            self.name = name
        except AttributeError:
            # Name is already a metadata of the ths, so we don't bother with it.
            pass

        self._reader = reader
        self._trace_klass = _metaclass.ClassWithMetadatas.__new__(
            cls=_metaclass.ClassWithMetadatas,
            name=f"Trace{reader.__class__.__name__}",
            bases=(trace.Trace, TraceDeprecated),
            namespace={},
            metadatas_keys=reader.metadatas_keys,
        )
        self._traces = None
        self._metadatas = None
        self._headers = None
        self._samples = None
        self._is_initialized = True

    def __str__(self):
        self_part = self[:3]
        r = f'Trace Header Set:\n{"Name":.<17}: {self.name}\n{"Reader":.<17}: {self._reader}\n'
        for k in self.metadatas.keys():
            r += f'{k:.<17}: {self_part.metadatas.get(k).dtype}\n'
        return r

    def __repr__(self):
        return str(self)

    def __len__(self):
        return len(self.traces)

    def __getattr__(self, name):
        try:
            return self.metadatas[name]
        except KeyError:
            return super().__getattr__(name)

    def __setattr__(self, name, value):
        if name not in self.__dir__() and self._is_initialized:
            self.metadatas  # Metadata needs to be instantiated before adding a in-memory metadata.
            self._metadatas[name] = value
        return super().__setattr__(name, value)

    def _transpose_slice(self, key):
        start = key.start
        stop = key.stop
        if start and start < 0:
            start = 0 if start < -len(self) else len(self) + start
        if stop and stop < 0:
            stop = 0 if stop < -len(self) else len(self) + stop
        elif stop and stop > len(self):
            stop = None
        return slice(start, stop, key.step)

    def __getitem__(self, key):
        """Implement slicing of :class:`TraceHeaderSet` instance.

        Args:
            key (int, slice, list): slice or item to return

        Returns:
            (:obj:`Trace`) if ``key`` is integer
            (:obj:`TraceHeaderSet`) new instance otherwise

        """
        if isinstance(key, (int, _np.integer)):
            if isinstance(key, _np.integer):
                key = key.item()
            trace = self.traces[key]
            if self._metadatas:
                trace._metadatas = self._metadatas._copy_with_cache(key=key)
            return trace
        elif isinstance(key, (slice, list, _np.ndarray)):
            if isinstance(key, _np.ndarray):
                if key.ndim > 1 or key.dtype.kind not in ['i', 'u']:
                    raise IndexError(f'Cannot slice TraceHeaderSet with numpy array of dim {key.ndim} and kind {key.dtype.kind}.')
            if not isinstance(key, slice) and max(key) > len(self):
                raise IndexError(f'Cannot index TraceHeaderSet of length {len(self)} with index {max(key)}.')

            if isinstance(key, slice):
                key = self._transpose_slice(key)

            new_ths = type(self).__new__(type(self))
            reader = self._reader[key]
            new_ths.__init__(name=self.name, reader=reader)
            if self._metadatas:
                new_ths._metadatas = self._metadatas._copy_with_cache(key=key, reader=reader)
            return new_ths
        else:
            if isinstance(key, tuple):
                raise IndexError('Too many indices for TraceHeaderSet.')
            raise IndexError(
                "Only integers, slices (':'), 1 dimension numpy array and lists ([1, 10, 5]) are valid indices for TraceHeaderSet."
            )

    def __iter__(self):
        for trc in self.traces:
            yield trc

    @property
    def traces(self):
        """Provide a list of :class:`Trace` instances from this trace header set."""
        if not self._traces:
            self._traces = TracesContainer(trace_klass=self._trace_klass, reader=self._reader)
        return self._traces

    @property
    def metadatas(self):
        """Provide a :class:`Metadatas` mapping-like object. Provides metadatas values for this trace header set."""
        if not self._metadatas:
            self._metadatas = metadatas.Metadatas(reader=self._reader)
        return self._metadatas

    @property
    def headers(self):
        if not self._headers:
            self._headers = headers.Headers(reader=self._reader)
        return self._headers

    @property
    def samples(self):
        """Provide a 2d :class:`Samples` instance ndarray-like of shape (number of traces, size of traces)."""
        if self._samples is None:
            self._samples = samples.Samples(reader=self._reader)
        return self._samples

    @property
    def filename(self):
        """If relevant, return the filename of the file containing the data, else return None."""
        try:
            return self._reader._filename
        except AttributeError:
            return None

    def split(self, part_size):
        """Return an iterable of :class:`TraceHeaderSet` slices of at least length ``part_size``.

        Args:
            part_size (int): number of traces per slice

        Returns:
            (:obj:`SplittedTraceHeaderSetIterable`) An iterable :class:`TraceHeaderSet` objects.

        """
        class SplittedTraceHeaderSetIterable:
            """Provide iterable around parts of equal size of a given :class:`TraceHeaderSet`."""

            def __init__(self, ths, part_size):
                self._ths = ths
                self._slices = [
                    slice(start * part_size, (start + 1) * part_size, 1)
                    for start in range(len(ths) // part_size)
                ]
                if len(ths) % part_size != 0:
                    self._slices.append(
                        slice(len(ths) // part_size * part_size, None, 1)
                    )

            def __iter__(self):
                for sl in self._slices:
                    yield self._ths[sl]

            def __getitem__(self, key):
                return self._ths[self._slices[key]]

            def __len__(self):
                return len(self._slices)

        if not isinstance(part_size, int) or part_size <= 0:
            raise TypeError("part_size must be a positive integer.")
        return SplittedTraceHeaderSetIterable(self, part_size)

    def filter(self, filter_function):
        """Build a new :class:`TraceHeaderSet` instance based on traces passing the ``filter_function`` condition.

        Args:
            filter_function (callable): function or lambda condition.
                Must take a trace as argument and return True or False.
        Returns:
            (:obj:`TraceHeaderSet`) a new instance with only selected traces.

        """
        keys_to_keep = [idx for idx, trc in enumerate(self) if filter_function(trc)]
        return self[keys_to_keep]

    def __reduce__(self):
        return build_trace_header_set, (self._reader, self.name)


def build_trace_header_set(reader: abstract_reader.AbstractReader, name: str) -> TraceHeaderSet:
    """Factory function which provides easy instantiation of :class:`TraceHeaderSet` instance.

    The function first creates a class, inheriting from :class:`TraceHeaderSet`,
    suited to the format instance provided, instantiate a trace header set and returns it.

    Warning:
        End user should not use directly this function, but factory functions suited to their concrete traces file format, provided by format developers.

    Args:
        reader: instance of a format reader class inheriting from :class:`AbstractReader`
        name: trace header set name

    Raises:
        TypeError: if ``reader`` does not implement :class:`AbstractReader`.

    Returns:
        :obj:`TraceHeaderSet`: trace header set instance.

    """
    if not isinstance(reader, abstract_reader.AbstractReader):
        raise TypeError(
            "reader must be a subclass of {f}".format(
                f=abstract_reader.AbstractReader.__name__
            )
        )
    ths_cls = _metaclass.ClassWithMetadatas.__new__(
        cls=_metaclass.ClassWithMetadatas,
        name=f"TraceHeaderSetFor{reader.__class__.__name__}",
        bases=(TraceHeaderSet, TraceHeaderSetDeprecated),
        namespace={},
        metadatas_keys=reader.metadatas_keys,
    )
    return ths_cls(name=name, reader=reader)



import collections
from . import abstract_reader


class Metadatas(collections.abc.Mapping):
    """Provides a dict-like object of traces metadatas.

    Each metadata value is either a value, if the `Metadata` instance wraps one trace,
    or an array of values, if the :class:`Metadata` wraps a trace set metadatas.

    """

    def __init__(self, reader: abstract_reader.AbstractReader, trace_id=None):
        if not isinstance(reader, abstract_reader.AbstractReader):
            raise TypeError(
                "reader must be a subclass of {f}".format(
                    f=abstract_reader.AbstractReader.__name__
                )
            )
        self._keys = reader.metadatas_keys
        self._trace_id = trace_id
        self._reader = reader
        self._cache = {}

    def __getitem__(self, key):
        if key not in self._keys:
            raise KeyError("Metadata with key {k} missing".format(k=key))
        if key not in self._cache:
            self._cache[key] = self._reader.fetch_metadatas(key=key, trace_id=self._trace_id)
        return self._cache[key]

    def __setitem__(self, key, value):
        self._cache[key] = value
        self._keys = list(self._keys) + [key]

    def is_trace(self):
        return self._trace_id is not None

    def __len__(self):
        return len(self._keys)

    def __iter__(self):
        return iter(self._keys)

    def __repr__(self):
        return f'{self._keys}-{self._reader}'

    def __str__(self):
        return repr(self)

    def _copy_with_cache(self, key, reader=None):
        if not reader:
            m = Metadatas(reader=self._reader, trace_id=key)
        else:
            m = Metadatas(reader=reader, trace_id=None)
        m._keys = self._keys
        m._cache = {
            k: v[key] for k, v in self._cache.items()
        }
        return m



import collections
from . import abstract_reader
import numpy as _np


class Headers(collections.abc.Mapping):
    """Provide a dict-like object of trace header set global metadata values."""

    def __init__(self, reader: abstract_reader.AbstractReader):
        if not isinstance(reader, abstract_reader.AbstractReader):
            raise TypeError(
                "reader must be a subclass of {f}".format(
                    f=abstract_reader.AbstractReader.__name__
                )
            )
        self._reader = reader
        self._keys = reader.headers_keys

    def __getitem__(self, key):
        if not isinstance(key, str):
            raise TypeError(f'key must be a str, not {type(key)}.')
        if key not in self._keys:
            raise KeyError(f'Header with key {key} missing')
        return self._reader.fetch_header(key=key)

    def __len__(self):
        return len(self._keys)

    def __iter__(self):
        return iter(self._keys)

    def __eq__(self, other):
        if not isinstance(other, collections.abc.Mapping):
            raise NotImplementedError(f'Cant compare Headers type with {type(other)} type.')
        if sorted(list(other.keys())) != sorted(list(self.keys())):
            return False
        for k, v in self.items():
            if isinstance(v, _np.ndarray):
                if not _np.array_equal(v, other[k]):
                    return False
            elif v != other[k]:
                return False
        return True

    def __str__(self):
        return str(list(self._keys))

    def __repr__(self):
        return str(self)



from .context import estraces  # noqa
import pytest
import numpy as np
from estraces.formats import random_reader
import cloudpickle


SMALL_SIZE = 10
BIG_SIZE = 9999999


@pytest.fixture
def rand(request):
    return random_reader.read_ths_from_random(**request.param)


@pytest.mark.parametrize('rand', [{'nb_traces': 0, 'trace_length': 0}], indirect=True)
def test_instanciate(rand):
    assert isinstance(rand, estraces.TraceHeaderSet)


@pytest.mark.parametrize('rand', [{'nb_traces': 0, 'trace_length': 0}], indirect=True)
def test_empty(rand):
    assert rand._reader.nb_traces == 0
    assert len(rand) == 0
    assert rand._reader.trace_length == 0


@pytest.mark.parametrize('rand', [{'nb_traces': SMALL_SIZE, 'trace_length': SMALL_SIZE}], indirect=True)
def test_small(rand):
    assert rand._reader.nb_traces == SMALL_SIZE
    assert len(rand) == SMALL_SIZE
    assert rand._reader.trace_length == SMALL_SIZE


@pytest.mark.parametrize('rand', [{'nb_traces': BIG_SIZE, 'trace_length': BIG_SIZE}], indirect=True)
def test_big(rand):
    assert rand._reader.nb_traces == BIG_SIZE
    assert len(rand) == BIG_SIZE
    assert rand._reader.trace_length == BIG_SIZE


def test_kwargtype_plainttext():
    rand1 = random_reader.read_ths_from_random(nb_traces=SMALL_SIZE,
                                               trace_length=SMALL_SIZE,
                                               plaintext=(16, 'uint8'))
    assert 'plaintext' in rand1.metadatas
    assert len(rand1.metadatas['plaintext'][0].tolist()) == 16
    assert isinstance(rand1.metadatas['plaintext'][0].tolist()[0], int)
    assert rand1.metadatas['plaintext'][0].dtype == 'uint8'


def test_kwargtype_shape():
    rand1 = random_reader.read_ths_from_random(nb_traces=SMALL_SIZE,
                                               trace_length=SMALL_SIZE,
                                               plaintext=(16, 'uint8'))
    assert rand1.metadatas['plaintext'].shape == (SMALL_SIZE, 16)


def test_missing_params_raise_type_error():
    samples = np.random.rand(10, 10)
    plaintext = (16, 'uint8')
    with pytest.raises(TypeError):
        random_reader.read_ths_from_random()
    with pytest.raises(TypeError):
        random_reader.read_ths_from_random(plaintext=plaintext)
    with pytest.raises(TypeError):
        random_reader.read_ths_from_random(samples[:10], plaintext=plaintext)
    with pytest.raises(TypeError):
        random_reader.read_ths_from_random(nb_traces=10, plaintext=plaintext)
    with pytest.raises(TypeError):
        random_reader.read_ths_from_random(trace_length=10, plaintext=plaintext)
    with pytest.raises(TypeError):
        random_reader.read_ths_from_random(trace_length='4445', plaintext=plaintext)


def test_incompatible_shapes():
    rand1 = random_reader.read_ths_from_random(nb_traces=SMALL_SIZE, trace_length=SMALL_SIZE, plaintext=(16, 'uint8'))
    assert rand1.samples[:].shape == (SMALL_SIZE, SMALL_SIZE)


def test_compatible_shapes_with_indexing():
    rand1 = random_reader.read_ths_from_random(nb_traces=SMALL_SIZE, trace_length=SMALL_SIZE, plaintext=(16, 'uint8'))
    assert rand1[:5].samples[:].shape == (5, SMALL_SIZE)


def test_incompatible_dtype():
    rand1 = random_reader.read_ths_from_random(nb_traces=SMALL_SIZE, trace_length=SMALL_SIZE, dtype='int16', plaintext=(16, 'uint8'))
    assert not rand1.samples[:].dtype == 'int8'


def test_compatible_default_dtype():
    rand1 = random_reader.read_ths_from_random(nb_traces=SMALL_SIZE, trace_length=SMALL_SIZE, plaintext=(16, 'uint8'))
    assert rand1.samples[:].dtype == 'int8'


def test_compatible_dtype():
    rand1 = random_reader.read_ths_from_random(nb_traces=SMALL_SIZE, trace_length=SMALL_SIZE, dtype='int16', plaintext=(16, 'uint8'))
    assert rand1.samples[:].dtype == 'int16'


@pytest.mark.parametrize('rand', [{'nb_traces': SMALL_SIZE, 'trace_length': SMALL_SIZE}], indirect=True)
def test_new_rand(rand):
    rand2 = random_reader.read_ths_from_random(nb_traces=SMALL_SIZE, trace_length=SMALL_SIZE)
    assert not np.array_equal(rand.samples[:], rand2.samples[:])


def test_indexing():
    rand = random_reader.read_ths_from_random(100, 1000, plaintext=(16, 'uint8'))
    assert len(rand[[3, 4, 5]]) == 3
    assert len(rand[np.array([10, 20, 15])]) == 3
    assert len(rand[10:20]) == 10
    assert len(rand[10:20:2]) == 5
    assert len(rand[90:]) == 10
    assert len(rand[:10]) == 10
    assert len(rand[:]) == 100


def test_indexing_2():
    rand = random_reader.read_ths_from_random(100, 1000, plaintext=(16, 'uint8'))
    assert rand._reader.fetch_samples(slice(None), slice(None)).shape == (100, 1000)
    assert rand[0:2].samples[:].shape == (2, 1000)
    assert rand[0:2].samples[:, ::2].shape == (2, 500)
    assert rand[0:2].samples[:, 100:].shape == (2, 900)
    assert rand.samples[:2, 100:].shape == (2, 900)
    assert rand[0:2].samples[:, 100:200].shape == (2, 100)


def test_multiple_indexing():
    rand = random_reader.read_ths_from_random(100, 1000, plaintext=(16, 'uint8'))
    sub1 = rand[:50]
    sub2 = sub1[40:60]
    assert len(sub1) == 50
    assert len(sub2) == 10
    assert sub1.samples[:].shape == (50, 1000)
    assert sub1.plaintext[:].shape == (50, 16)
    assert sub2.samples[:].shape == (10, 1000)
    assert sub2.plaintext[:].shape == (10, 16)


SEED = 101


@pytest.mark.parametrize('rand', [{'nb_traces': 2, 'trace_length': 2, 'seed': SEED}], indirect=True)
def test_fixed_seed(rand):
    sample1 = rand.samples[:]
    rand2 = random_reader.read_ths_from_random(nb_traces=2, trace_length=2, seed=SEED)
    assert np.array_equal(sample1, rand2.samples[:])


def test_serialization():
    ths = random_reader.read_ths_from_random(100, 1000, plaintext=(16, 'uint8'))
    dumped = cloudpickle.dumps(ths)
    loaded_ths = cloudpickle.loads(dumped)
    assert loaded_ths.samples[:].shape == ths.samples[:].shape
    assert loaded_ths.samples[:].dtype == ths.samples[:].dtype
    assert loaded_ths.plaintext[:].shape == ths.plaintext[:].shape
    assert loaded_ths.plaintext[:].dtype == ths.plaintext[:].dtype


# see https://gitlab.com/eshard/estraces/-/issues/45
def test_filename_property():
    ths = random_reader.read_ths_from_random(100, 1000, plaintext=(16, 'uint8'))
    assert ths.filename is None



import numpy as _np

from ..traces.abstract_reader import AbstractReader
from ..traces.trace_header_set import build_trace_header_set


def read_ths_from_ram(samples, **kwargs):
    """Build and return a :class:`TraceHeaderSet` instance from arrays in RAM memory.

    Args:
        samples (ndarray): traces samples
        **kwargs (ndarray): metadata as ndarray

    Returns:
        (:obj:`TraceHeaderSet`)

    """
    return build_trace_header_set(
        reader=RAMReader(samples=samples, **kwargs),
        name='RAM Format THS'
    )


class RAMReader(AbstractReader):

    def __init__(self, samples, headers={}, **kwargs):
        self._test_nd_arrays(samples, kwargs)
        self._test_shapes_compatibility(samples, kwargs)

        self._samples = samples
        self._kwargs_dict = kwargs
        self._headers = {k: v for k, v in headers.items()}
        self._size, self._trace_size = samples.shape

    @staticmethod
    def _test_nd_arrays(samples, kwargs_dict):
        if not isinstance(samples, _np.ndarray):
            raise TypeError(f'`samples` argument must be a {_np.ndarray}')
        if samples.ndim != 2:
            raise TypeError('`samples` argument must be a 2-dimensions ndarray')
        for key in kwargs_dict.keys():
            if not isinstance(kwargs_dict[key], _np.ndarray):
                raise TypeError(f'`{key}` argument must be a {_np.ndarray}')

    @staticmethod
    def _test_shapes_compatibility(samples, kwargs_dict):
        samples_shape = samples.shape
        for key in kwargs_dict.keys():
            if kwargs_dict[key].shape[0] != samples_shape[0]:
                raise ValueError(f'Incompatible shapes: samples {samples_shape} and {key} {kwargs_dict[key].shape}')

    def fetch_metadatas(self, key, trace_id):
        if trace_id is not None:
            return self._kwargs_dict[key][trace_id]
        else:
            return self._kwargs_dict[key]

    def fetch_header(self, key):
        return self._headers[key]

    def fetch_samples(self, traces, frame):
        if isinstance(traces, int):
            traces = [traces]
        if isinstance(frame, int):
            frame = [frame]
        if frame is not None:
            return self._samples[traces][:, frame]
        else:
            return self._samples[traces]

    def __getitem__(self, key):
        super().__getitem__(key)
        if isinstance(key, int):
            key = [key]
        new_samples = self._samples[key]
        new_kwargs = {item[0]: item[1][key] for item in self._kwargs_dict.items()}
        new_reader = RAMReader(new_samples, headers=self._headers, **new_kwargs)
        return new_reader

    @property
    def metadatas_keys(self):
        """Provide a list or views of the metadatas keys available."""
        return self._kwargs_dict.keys()

    @property
    def headers_keys(self):
        """Provide a list or views of the headers keys available."""
        return self._headers.keys()

    def get_trace_size(self, trace_id):
        """Provide the size of trace trace_id."""
        return self._trace_size

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return f'RAM reader with {len(self)} traces. Samples shape: {self._samples.shape} and metadatas: {list(self._kwargs_dict.keys())}'
